// SPDX-FileCopyrightText: 2021 GNOME Foundation
// SPDX-License-Identifier: LGPL-2.1-or-later

// A map between namespaces and base URLs for their online documentation
baseURLs = [
    [ 'GLib', 'https://docs.gtk.org/glib/' ],
    [ 'GObject', 'https://docs.gtk.org/gobject/' ],
    [ 'Gio', 'https://docs.gtk.org/gio/' ],
    [ 'Atk', 'https://docs.gtk.org/atk/' ],
    [ 'Gdk', 'https://docs.gtk.org/gdk3/' ],
    [ 'Gtk', 'https://docs.gtk.org/gtk3/' ],
    [ 'Pango', 'https://docs.gtk.org/Pango/' ],
    [ 'PangoCairo', 'https://docs.gtk.org/PangoCairo/' ],
    [ 'GdkPixbuf', 'https://docs.gtk.org/gdk-pixbuf/' ],
]
