/* meta-cogl-utils.h
 *
 * Copyright 2020 Georges Basile Stavracas Neto <georges.stavracas@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 */

#ifndef META_COGL_UTILS_H
#define META_COGL_UTILS_H

#include "cogl/cogl.h"

G_BEGIN_DECLS

gboolean
meta_cogl_pixel_format_from_drm_format (uint32_t               drm_format,
                                        CoglPixelFormat       *out_format,
                                        CoglTextureComponents *out_components);

G_END_DECLS

#endif /* META_COGL_UTILS_H */
